from typing import Optional
from enum import Enum, auto


class CNFLogicConnective(Enum):
    """The propositional logic connectives a formula in CNF can contain."""
    NOT = auto()
    AND = auto()
    OR = auto()


class TextCNFNotation(Enum):
    """Possible notations of a CNF formula.

    Five different notations are available:
        Logical symbols:
            (A) ∧ (¬B ∨ C) ∧ ...
        Textual symbols:
            (A) and (not B or C) and ...
        Java symbols:
            (A) && (!B || C) && ...
        Java Short symbols:
            (A) & (!B | C) && ...
        Short symbols:
            (A) & (-B | C) & ...
    """
    LOGICAL = {CNFLogicConnective.NOT: '¬',
               CNFLogicConnective.AND: '∧',
               CNFLogicConnective.OR: '∨'}
    TEXTUAL = {CNFLogicConnective.NOT: 'not',
               CNFLogicConnective.AND: 'and',
               CNFLogicConnective.OR: 'or'}
    JAVA = {CNFLogicConnective.NOT: '!',
            CNFLogicConnective.AND: '&&',
            CNFLogicConnective.OR: '||'}
    JAVA_SHORT = {CNFLogicConnective.NOT: '!',
                  CNFLogicConnective.AND: '&',
                  CNFLogicConnective.OR: '|'}
    SHORT = {CNFLogicConnective.NOT: '-',
             CNFLogicConnective.AND: '&',
             CNFLogicConnective.OR: '|'}


class TextCNFModel():
    """Textual representation of a conjunctive normal form (CNF) formula.

    A CNF formula (or clausal normal form) is a conjunction of one or more clauses,
    where a clause is a disjunction of literals.
    """

    def __init__(self) -> None:
        self._cnf_formula: Optional[str] = None
        self._cnf_notation: Optional[TextCNFNotation] = None
        self._variables: list[str] = []  # list of str with variables' names

    def from_textual_cnf(self, cnf_formula: str) -> None:
        self._cnf_formula = cnf_formula
        self._cnf_notation = identify_notation(cnf_formula)
        self._variables = extract_variables(cnf_formula)

    def from_textual_cnf_file(self, filepath: str) -> None:
        """This method reads any of the available textual notations,
        but only one notation at the same time,
        so the .txt file should include only one of the possible notations in a single line.
        """
        with open(filepath, 'r', encoding='utf-8') as file:
            self.from_textual_cnf(file.readline())

    def write_textual_cnf_file(self, filepath: str,
                               syntax: TextCNFNotation = TextCNFNotation.JAVA_SHORT) -> None:
        """Write the textual CNF formula as a string in a file.

        Default syntax is TextCNFNotation.JAVA_SHORT: (A) & (!B | C) && ...
        """
        cnf_formula = self.get_textual_cnf_formula(syntax)
        with open(filepath, 'w+', encoding='utf-8') as file:
            file.write(cnf_formula)

    def get_textual_cnf_notation(self) -> TextCNFNotation:
        """Return the notation used for the CNF formula."""
        if self._cnf_formula is None:
            raise Exception("CNF Model not initialized. Use a `from_` method first.")
        assert self._cnf_notation is not None
        return self._cnf_notation

    def get_textual_cnf_formula(self,
                                syntax: Optional[TextCNFNotation] = None) -> str:
        """Return the CNF formula in the specified notation syntax.

        Default syntax is TextCNFNotation.JAVA_SHORT: (A) & (!B | C) && ...
        """
        if self._cnf_formula is None:
            raise Exception("CNF Model not initialized. Use a `from_` method first.")
        assert self._cnf_notation is not None

        cnf_formula = self._cnf_formula
        cnf_notation = self._cnf_notation

        if syntax is not None and syntax == cnf_notation:
            return cnf_formula
        if syntax is None:
            syntax = cnf_notation

        # Translate AND operators
        symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.AND] + ' '
        new_symbol = ' ' + syntax.value[CNFLogicConnective.AND] + ' '
        cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

        # Translate OR operators
        symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.OR] + ' '
        new_symbol = ' ' + syntax.value[CNFLogicConnective.OR] + ' '
        cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

        # Translate NOT operators
        # This is more complex because the symbol may be part of a feature's name
        if cnf_notation == TextCNFNotation.TEXTUAL:
            symbol_pattern = cnf_notation.value[CNFLogicConnective.NOT] + ' '
            new_symbol = syntax.value[CNFLogicConnective.NOT]
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
        elif syntax == TextCNFNotation.TEXTUAL:
            symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.NOT]
            new_symbol = ' ' + syntax.value[CNFLogicConnective.NOT] + ' '
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

            symbol_pattern = '(' + cnf_notation.value[CNFLogicConnective.NOT]
            new_symbol = '(' + syntax.value[CNFLogicConnective.NOT] + ' '
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
        else:
            symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.NOT]
            new_symbol = ' ' + syntax.value[CNFLogicConnective.NOT]
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

            symbol_pattern = '(' + cnf_notation.value[CNFLogicConnective.NOT]
            new_symbol = '(' + syntax.value[CNFLogicConnective.NOT]
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
        return cnf_formula

    def get_variables(self) -> list[str]:
        """Return the list of variables' names in the CNF formula."""
        if self._cnf_formula is None:
            raise Exception("CNF Model not initialized. Use a `from_` method first.")
        return self._variables


def identify_notation(cnf_formula: str) -> TextCNFNotation:
    """Return the notation used by the given CNF formula.

    Default TextCNFNotation.JAVA.
    """
    notation = check_unary_connective(cnf_formula)
    if notation is None or notation in (TextCNFNotation.JAVA, TextCNFNotation.JAVA_SHORT):
        notation = check_binary_connective(cnf_formula)
    if notation is None:
        notation = TextCNFNotation.JAVA_SHORT
    return notation


def check_unary_connective(cnf_formula: str) -> Optional[TextCNFNotation]:
    symbol = TextCNFNotation.LOGICAL.value[CNFLogicConnective.NOT]
    if ' ' + symbol in cnf_formula or '(' + symbol in cnf_formula:
        return TextCNFNotation.LOGICAL

    symbol = TextCNFNotation.SHORT.value[CNFLogicConnective.NOT]
    if ' ' + symbol in cnf_formula or '(' + symbol in cnf_formula:
        return TextCNFNotation.SHORT

    symbol = TextCNFNotation.TEXTUAL.value[CNFLogicConnective.NOT]
    if ' ' + symbol + ' ' in cnf_formula or '(' + symbol + ' ' in cnf_formula:
        return TextCNFNotation.TEXTUAL

    symbol = TextCNFNotation.JAVA.value[CNFLogicConnective.NOT]  # JAVA or JAVA_SHORT
    if ' ' + symbol in cnf_formula or '(' + symbol in cnf_formula:
        return TextCNFNotation.JAVA

    return None


def check_binary_connective(cnf_formula: str) -> Optional[TextCNFNotation]:
    for notation in TextCNFNotation:
        for connective in notation.value.keys():
            if connective != CNFLogicConnective.NOT:
                symbol = notation.value[connective]
                symbol_pattern = ' ' + symbol + ' '
                if symbol_pattern in cnf_formula:
                    return notation
    return None


def extract_variables(cnf_formula: str) -> list[str]:
    """Return the list of variables' names of the CNF formula."""
    variables = set()
    cnf_notation = identify_notation(cnf_formula)

    # Remove initial and final parenthesis
    and_symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.AND] + ' '
    clauses = list(map(lambda c: c[1:len(c) - 1], cnf_formula.split(and_symbol_pattern)))

    # Remove final parenthesis of last clause (because of the possible end of line: '\n')
    if ')' in clauses[len(clauses) - 1]:
        clauses[len(clauses) - 1] = clauses[len(clauses) - 1][:-1]

    for clause in clauses:
        tokens = clause.split(' ')
        tokens = list(filter(lambda t: t != cnf_notation.value[CNFLogicConnective.OR], tokens))
        for feature in tokens:
            if feature == cnf_notation.value[CNFLogicConnective.NOT]:
                continue
            if feature.startswith(cnf_notation.value[CNFLogicConnective.NOT]):
                variables.add(feature.replace(cnf_notation.value[CNFLogicConnective.NOT], '', 1))
            else:
                variables.add(feature)
    return list(variables)



from pysat.formula import CNF

from flamapy.core.models import VariabilityModel


class PySATModel(VariabilityModel):

    @staticmethod
    def get_extension() -> str:
        return 'pysat'

    def __init__(self) -> None:
        #self.r_cnf = CNF()  # ToDo: This should be avoid
        #self.ctc_cnf = CNF()  # ToDo: This should be avoid
        self._cnf = CNF()
        self.variables: dict[str, int] = {}  # feature's name -> id
        self.features: dict[int, str] = {}  # id -> feature's name

    def add_clause(self, clause: list[int]) -> None:
        #self.ctc_cnf.append(clause)
        self._cnf.append(clause)

    def get_all_clauses(self) -> CNF:
        #clauses = CNF()
        #clauses.extend(self.r_cnf.clauses)
        #clauses.extend(self.ctc_cnf.clauses)
        #return clauses
        return self._cnf



from pysat.solvers import Glucose3

from flamapy.core.operations import ValidConfiguration
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration


from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3ValidConfiguration(ValidConfiguration):

    def __init__(self) -> None:
        self.result = False
        self.configuration = Configuration({})

    def is_valid(self) -> bool:
        return self.result

    def get_result(self) -> bool:
        return self.is_valid()

    def set_configuration(self, configuration: Configuration) -> None:
        self.configuration = configuration

    def execute(self, model: PySATModel) -> 'Glucose3ValidConfiguration':
        glucose = Glucose3()

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            glucose.add_clause(clause)  # añadimos la constraint

        assumptions = []
        for feat in self.configuration.elements.items():
            if feat[1]:
                assumptions.append(model.variables[feat[0].name])
            elif not feat[1]:
                assumptions.append(-model.variables[feat[0].name])

        self.result = glucose.solve(assumptions=assumptions)
        glucose.delete()
        return self



from pysat.solvers import Glucose3

from flamapy.core.operations import Valid

from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3Valid(Valid):

    def __init__(self) -> None:
        self.result = False

    def is_valid(self) -> bool:
        return self.result

    def get_result(self) -> bool:
        return self.is_valid()

    def execute(self, model: PySATModel) -> 'Glucose3Valid':
        glucose = Glucose3()
        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            glucose.add_clause(clause)  # añadimos la constraint
        self.result = glucose.solve()
        glucose.delete()
        return self



from typing import Any

from pysat.solvers import Glucose3

from flamapy.core.operations import DeadFeatures
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3DeadFeatures(DeadFeatures):

    def __init__(self) -> None:
        self.dead_features: list[list[Any]] = []

    def get_dead_features(self) -> list[list[Any]]:
        return self.dead_features

    def get_result(self) -> list[list[Any]]:
        return self.get_dead_features()

    def execute(self, model: PySATModel) -> 'Glucose3DeadFeatures':
        glucose = Glucose3()

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            glucose.add_clause(clause)  # añadimos la constraint

        dead_features = []
        for variable in model.variables.items():
            if not glucose.solve(assumptions=[variable[1]]):
                dead_features.append(variable[0])
        self.dead_features = dead_features
        glucose.delete()
        return self



from flamapy.core.operations import Commonality
from flamapy.core.models import VariabilityModel
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.metamodels.pysat_metamodel.operations.glucose3_products import Glucose3Products


class Glucose3Commonality(Commonality):
    def __init__(self) -> None:
        self.commonality: float = 0
        self.configuration = Configuration({})

    def set_configuration(self, configuration: Configuration) -> None:
        self.configuration = configuration

    def get_commonality(self) -> float:
        return self.commonality

    def get_result(self) -> float:
        return self.get_commonality()

    def execute(self, model: VariabilityModel) -> 'Glucose3Commonality':
        glucose3products = Glucose3Products()
        glucose3products.execute(model)

        products = glucose3products.get_result()

        feature = list(self.configuration.elements.keys())[0]

        count = 0
        for product in products:
            count = count + \
                1 if feature.name in product else count

        self.commonality = count / len(products)

        return self



from .glucose3_valid import Glucose3Valid
from .glucose3_valid_configuration import Glucose3ValidConfiguration
from .glucose3_valid_product import Glucose3ValidProduct
from .glucose3_products import Glucose3Products
from .glucose3_products_number import Glucose3ProductsNumber
from .glucose3_commonality import Glucose3Commonality
from .glucose3_filter import Glucose3Filter
from .glucose3_core_features import Glucose3CoreFeatures
from .glucose3_dead_features import Glucose3DeadFeatures
from .glucose3_false_optional_features import Glucose3FalseOptionalFeatures
from .glucose3_error_detection import Glucose3ErrorDetection


__all__ = [
    'Glucose3Valid', 
    'Glucose3ValidConfiguration', 
    'Glucose3ValidProduct', 
    'Glucose3Products', 
    'Glucose3ProductsNumber', 
    'Glucose3Commonality', 
    'Glucose3Filter', 
    'Glucose3CoreFeatures', 
    'Glucose3DeadFeatures', 
    'Glucose3FalseOptionalFeatures', 
    'Glucose3ErrorDetection',  
]



from typing import Any

from pysat.solvers import Glucose3

from flamapy.core.operations import FalseOptionalFeatures
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.metamodels.fm_metamodel.models.feature_model import FeatureModel


class Glucose3FalseOptionalFeatures(FalseOptionalFeatures):

    def __init__(self, feature_model: FeatureModel) -> None:
        self.result: list[Any] = []
        self.feature_model = feature_model

    def execute(self, model: PySATModel) -> 'Glucose3FalseOptionalFeatures':
        self.result = get_false_optional_features(model, self.feature_model)
        return self

    def get_false_optional_features(self) -> list[list[Any]]:
        return self.get_result()

    def get_result(self) -> list[Any]:
        return self.result


def get_false_optional_features(sat_model: PySATModel, feature_model: FeatureModel) -> list[Any]:
    real_optional_features = [f for f in feature_model.get_features() 
                              if not f.is_root() and not f.is_mandatory()]

    result = []
    solver = Glucose3()
    for clause in sat_model.get_all_clauses():
        solver.add_clause(clause)

    for feature in real_optional_features:
        variable = sat_model.variables.get(feature.name)
        parent_variable = sat_model.variables.get(feature.get_parent().name)
        assert variable is not None
        satisfiable = solver.solve(assumptions=[parent_variable, -variable])
        if not satisfiable:
            result.append(feature.name)
    solver.delete()
    return result



from typing import Any

from pysat.solvers import Glucose3

from flamapy.core.operations import Products
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3Products(Products):

    def __init__(self) -> None:
        self.products: list[list[Any]] = []

    def get_products(self) -> list[list[Any]]:
        return self.products

    def get_result(self) -> list[list[Any]]:
        return self.get_products()

    def execute(self, model: PySATModel) -> 'Glucose3Products':
        glucose = Glucose3()

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            glucose.add_clause(clause)  # añadimos la constraint

        for solutions in glucose.enum_models():
            product = []
            for variable in solutions:
                if variable > 0:
                    product.append(model.features.get(variable))
            self.products.append(product)
        glucose.delete()
        return self



from typing import Any

from pysat.solvers import Glucose3

from flamapy.core.operations import CoreFeatures
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3CoreFeatures(CoreFeatures):

    def __init__(self) -> None:
        self.core_features: list[list[Any]] = []

    def get_core_features(self) -> list[list[Any]]:
        return self.core_features

    def get_result(self) -> list[list[Any]]:
        return self.get_core_features()

    def execute(self, model: PySATModel) -> 'Glucose3CoreFeatures':
        glucose = Glucose3()
        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            glucose.add_clause(clause)  # añadimos la constraint

        core_features = []
        if glucose.solve():
            for variable in model.variables.items():
                if not glucose.solve(assumptions=[-variable[1]]):
                    core_features.append(variable[0])

        self.core_features = core_features
        glucose.delete()
        return self
